#include "crack.h"
#undef VERBOSE
#undef LISTING

/* LOOK FOR CONFIGURATION OPTIONS IN "crack.h" */

/*
 * This program is copyright (c) Alec Muffett 1991 except for certain
 * portions of code ("fdes-crypt.c") copyright (c) Robert Baldwin, Icarus
 * Sparry and Alec Muffett.  The author(s) disclaims all responsibility or
 * liability with respect to it's usage or its effect upon hardware or
 * computer systems.  This software is in the public domain and is freely
 * redisrtibutable PROVIDED that this copyright notice remains intact.
 */

/*
 * crack-pwc.c - an optimised password cracker. (c) ADE Muffett, Dec 1990.
 * Totally rewritten from an 'standard' password cracker that had been
 * floating about on ther network. Apart from the entire re-write, to save my
 * sanity, this thing orders passwords to minimise calls to crypt(). Use
 * fcrypt() where you have it.
 */

struct ENTRY
{
    int done;
    char *filename;
    struct passwd passwd;
    struct ENTRY *across;
    struct ENTRY *next;
};

static struct ENTRY *root;

/*
 * DIDDY LITTLE USEFUL FUNCTIONS
 */

void
Log (fmt, a, b, c, d, e, f, g, h, i, j)
    char *fmt;
{
    long t;

    time (&t);
    printf ("pwc: %-15.15s ", ctime (&t) + 4);
    printf (fmt, a, b, c, d, e, f, g, h, i, j);
    fflush (stdout);
}
/*
 * STRUCTURE SPECIFIC FUNCTIONS
 */

char *
PWSkip (p)			/* jump the ':' in a pwent */
    register char *p;
{
    while (*p && *p != ':')
    {
	p++;
    }
    if (*p)
    {
	*p++ = '\0';
    }
    return (p);
}

struct ENTRY *
Parse (buffer)			/* break up input into a structure */
    register char *buffer;
{
    register char *p;
    register struct ENTRY *retval;

    retval = (struct ENTRY *) malloc (sizeof (struct ENTRY));
    retval -> next = retval -> across = NULL;
    retval -> done = 0;
    Trim (buffer);

    p = Clone (buffer);
    retval -> filename = p;
    p = PWSkip (p);
    retval -> passwd.pw_name = p;
    p = PWSkip (p);
    retval -> passwd.pw_passwd = p;
    p = PWSkip (p);
    retval -> passwd.pw_uid = 1;
    p = PWSkip (p);
    retval -> passwd.pw_gid = 1;
    p = PWSkip (p);
    retval -> passwd.pw_gecos = p;
    p = PWSkip (p);
    retval -> passwd.pw_dir = p;
    p = PWSkip (p);
    retval -> passwd.pw_shell = p;
    return (retval);
}
/*
 * START OF MODULES
 */

void
LoadData ()			/* load sorted entries into memory */
{
    char *ptr;
    char salt[2];
    char buffer[STRINGSIZE];
    int numlines;
    int numentries;
    register struct ENTRY *new_element;
    register struct ENTRY *current_line;

    numlines = 0;
    numentries = 0;
    current_line = NULL;

    while (fgets (buffer, STRINGSIZE, stdin))
    {
	if (!*buffer || isspace (*buffer))
	{
	    continue;
	}
	new_element = Parse (buffer);

	ptr = new_element -> passwd.pw_passwd;
	if (!ptr[0] || ptr[0] == '*')
	{
	    continue;
	}
	numentries++;

	if (ptr[0] == salt[0] && ptr[1] == salt[1])
	{
	    new_element -> across = current_line;
	    current_line = new_element;

	} else
	{
	    if (current_line)
	    {
		current_line -> next = root;
	    }
	    root = current_line;
	    current_line = new_element;
	    numlines++;
	    salt[0] = ptr[0];
	    salt[1] = ptr[1];
	}
    }

    if (current_line)		/* last one tends to hang about */
    {
	current_line -> next = root;
	root = current_line;
	numlines++;
    }
    Log ("Loaded %d password entries into %d salted lines.\n",
	 numentries, --numlines);

    return;
}
/*
 * PASSWORD CRACKING LOW LEVEL FUNCTIONS
 */

int
TryManyUsers (eptr, guess)	/* returns 0 if all done this line */
    register struct ENTRY *eptr;
    char *guess;
{
    register int retval;
    char guess_crypted[STRINGSIZE];

    if (eptr -> done && !eptr -> across)
    {
	return (0);
    }
    strcpy (guess_crypted, crypt (guess, eptr -> passwd.pw_passwd));

    retval = 0;

    while (eptr)
    {
#ifdef VERBOSE
	Log ("Trying '%s' on %s from line %s\n",
	     guess,
	     eptr -> passwd.pw_name,
	     eptr -> filename);
#endif
	if (!strcmp (guess_crypted, eptr -> passwd.pw_passwd))
	{
	    if (!eptr -> done)	/* haven't printed it before */
	    {
		Log ("Guessed %s (%s in %s) [%s] \n",
		     eptr -> passwd.pw_name,
		     eptr -> passwd.pw_shell,
		     eptr -> filename,
#ifdef CRACK_PRINT
		     guess
#else
		     ""
#endif				/* CRACK_PRINT */
		    );
	    }
	    eptr -> done = 1;
	}
	retval += (!eptr -> done);
	eptr = eptr -> across;
    }

    return (retval);
}

int
TryOneUser (eptr, guess)	/* returns non-null on guessed user */
    register struct ENTRY *eptr;
    register char *guess;
{
    if (!guess || !*guess || eptr -> done)
    {
	return (0);
    }
#ifdef VERBOSE
    Log ("Trying '%s' on %s from %s\n",
	 guess,
	 eptr -> passwd.pw_name,
	 eptr -> filename);
#endif
    if (strcmp (crypt (guess, eptr -> passwd.pw_passwd),
		eptr -> passwd.pw_passwd))
    {
	return (0);
    }
    eptr -> done = 1;
    Log ("Guessed %s (%s in %s) [%s] \n",
	 eptr -> passwd.pw_name,
	 eptr -> passwd.pw_shell,
	 eptr -> filename,
#ifdef CRACK_PRINT
	 guess
#else
	 ""
#endif				/* CRACK_PRINT */
	);

    return (1);
}
/*
 * TOP LEVEL CRACKING FUNCTION INTERFACE ROUTINE - SINGLE USER ONLY
 */

int
WordTry (entry_ptr, guess)
    register struct ENTRY *entry_ptr;
    register char *guess;
{
    int all_lower;
    int all_upper;
    register char *ptr;
    char guess2[STRINGSIZE];

    if (!guess[1])		/* avoid wasting time on initials */
    {
	return (0);
    }
    if (TryOneUser (entry_ptr, guess) ||
	TryOneUser (entry_ptr, Reverse (guess)))
    {
	return (1);
    }
    all_upper = all_lower = 1;
    ptr = guess;

    for (ptr = guess; *ptr; ptr++)
    {
	if (islower (*ptr))
	{
	    all_upper = 0;
	} else if (isupper (*ptr))
	{
	    all_lower = 0;
	}
    }

    if (!all_lower)
    {
	strcpy (guess2, Lowercase (guess));
	if (TryOneUser (entry_ptr, guess2) ||
	    TryOneUser (entry_ptr, Reverse (guess2)))
	{
	    return (1);
	}
    }
    if (!all_upper)
    {
	strcpy (guess2, Uppercase (guess));
	if (TryOneUser (entry_ptr, guess2) ||
	    TryOneUser (entry_ptr, Reverse (guess2)))
	{
	    return (1);
	}
    }
    return (0);
}
/*
 * END OF PASSWORD GUESSING LAYERS
 */

void
Pass1 ()
{
    int cracked;
    struct ENTRY *head;
    register char *ptr;
    register char *ptr2;
    char junk[STRINGSIZE];
    register struct ENTRY *this;

    Log ("starting pass1 password information\n");

    for (head = root; head; head = head -> next)
    {
	for (this = head; this; this = this -> across)
	{
#ifdef CRACK_USERNAME
	    /* username */
	    ptr = this -> passwd.pw_name;
	    if (WordTry (this, ptr))
	    {
		continue;
	    }
#endif				/* CRACK_USERNAME */

#ifdef CRACK_USERNAME2
	    /* usernameusername */
	    strcpy (junk, ptr);
	    strcat (junk, ptr);
	    if (WordTry (this, junk))
	    {
		continue;
	    }
#endif				/* CRACK_USERNAME2 */

#ifdef CRACK_GECOS
	    /* Gecos information field */
	    cracked = 0;
	    ptr = junk;
	    strcpy (junk, this -> passwd.pw_gecos);
	    if (*ptr == '-')	/* never seen this, but... */
	    {
		ptr++;
	    }
	    if (ptr2 = strchr (ptr, ';'))	/* trim off junk */
	    {
		*ptr2 = '\0';
	    }
	    if (ptr2 = strchr (ptr, ','))	/* trim off more junk */
	    {
		*ptr2 = '\0';
	    }
	    for (;;)
	    {
		if (ptr2 = strchr (ptr, ' '))
		{
		    *(ptr2++) = '\0';
		}
		if (WordTry (this, ptr))
		{
		    cracked++;
		    break;
		}
		if (ptr2)
		{
		    ptr = ptr2;
		    while (*ptr && isspace (*ptr))
		    {
			ptr++;
		    }
		} else
		{
		    break;
		}
	    }

	    if (cracked)
	    {
		continue;
	    }
#endif				/* CRACK_GECOS */
	}
    }
    return;
}

void
Pass2 (dictfile)
    char *dictfile;
{
    FILE *fp;
    int dictnum;
    register int i;
    char *dict[MAXDICTSIZE];
    char buffer[STRINGSIZE];
    register struct ENTRY *head;

    Log ("starting pass2 on dictionary %s\n", dictfile);

    if (!(fp = fopen (dictfile, "r")))
    {
	perror (dictfile);
	return;
    }
    for (dictnum = 0;
	 dictnum < MAXDICTSIZE && fgets (buffer, STRINGSIZE, fp);
	 dictnum++)
    {
	Trim (buffer);
	if (!*buffer)
	{
	    continue;
	}
	dict[dictnum] = Clone (buffer);
    }

    if (dictnum >= MAXDICTSIZE)
    {
	Log ("dictionary %s too big, truncated at %d words\n",
	     dictfile, MAXDICTSIZE);
    }
    fclose (fp);

    Log ("loaded %d words from %s\n", dictnum, dictfile);
    for (head = root; head; head = head -> next)
    {
	for (i = 0; i < dictnum; i++)
	{
	    if (!TryManyUsers (head, dict[i]))
	    {
		break;
	    }
	}
    }

    Log ("Freeing memory of %s\n", dictfile);
    while (dictnum >= 0)
    {
	if (dict[dictnum])
	{
	    free (dict[dictnum]);
	}
	dictnum--;
    }
}

int
main (argc, argv)
    int argc;
    char *argv[];
{
    int i;
    FILE *fp;
    char *file;
    static char die[] = "die";

    if (argc < 2)
    {
	fprintf (stderr, "Usage:\t%s dictfile [dictfile ...]\n", argv[0]);
	exit (1);
    }
    if (!(fp = fopen (die, "w")))
    {
	perror (die);
	exit (2);
    }
    fprintf (fp, "#!/bin/sh\nkill -9 %d\n", getpid ());
    fclose (fp);
    chmod (die, 0700);

#ifdef FCRYPT
    init_des ();
#endif

    LoadData ();

    Pass1 ();

    for (i = 1; argv[i]; i++)
    {
	Pass2 (argv[i]);
    }

    Log ("Finished.\n");
    return (0);
}
